﻿/*
 * Copyright 2010 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

using System;
using System.Collections.Generic;
using System.Threading;
using System.Globalization;

using NUnit.Framework;

namespace Vajhoej.Record.Test
{
    [TestFixture]
    public class TestBCDUtil
    {
        [SetUp]
        protected void Init() {
            Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;
        }
        [Test]
        public void TestPackedBCD()
        {
            byte[] b1 = { 0x12, 0x7C };
            decimal v1 = BCDUtil.DecodePackedBCD(b1, 0);
            Assert.AreEqual(decimal.Parse("127"), v1, "v1");
            decimal v2 = decimal.Parse("127");
            byte[] b2 = BCDUtil.EncodePackedBCD(v2, 0, 2);
            Assert.AreEqual(b1.Length, b2.Length, "b2 length");
            for(int i = 0; i < b2.Length; i++)
            {
                Assert.AreEqual(b1[i], b2[i], "b2 element " + i);
            }
            byte[] b3 = { 0x12, 0x34, 0x56, 0x7D };
            decimal v3 = BCDUtil.DecodePackedBCD(b3, 2);
            Assert.AreEqual(decimal.Parse("-12345.67"), v3, "v3");
            decimal v4 = decimal.Parse("-12345.67");
            byte[] b4 = BCDUtil.EncodePackedBCD(v4, 2, 4);
            Assert.AreEqual(b3.Length, b4.Length, "b4 length");
            for(int i = 0; i < b4.Length; i++)
            {
                Assert.AreEqual(b3[i], b4[i], "b4 element " + i);
            }
        }
        [Test]
        public void TestZonedBCD()
        {
            byte[] b1 = { (byte)0xF1, (byte)0xF2, (byte)0xD3 };
            decimal v1 = BCDUtil.DecodeZonedBCD(b1, BCDUtil.EBCDIC, 0);
            Assert.AreEqual(decimal.Parse("-123"), v1, "v1");
            decimal v2 = decimal.Parse("-123");
            byte[] b2 = BCDUtil.EncodeZonedBCD(v2, BCDUtil.EBCDIC, 0, 3);
            Assert.AreEqual(b1.Length, b2.Length, "b2 length");
            for(int i = 0; i < b2.Length; i++)
            {
                Assert.AreEqual(b1[i], b2[i], "b2 element " + i);
            }
            byte[] b3 = { (byte)0xF1, (byte)0xF2, (byte)0xF7, (byte)0xF9, (byte)0xF5, (byte)0xC0 };
            decimal v3 = BCDUtil.DecodeZonedBCD(b3, BCDUtil.EBCDIC, 2);
            Assert.AreEqual(decimal.Parse("1279.50"), v3, "v3");
            decimal v4 = decimal.Parse("1279.50");
            byte[] b4 = BCDUtil.EncodeZonedBCD(v4, BCDUtil.EBCDIC, 2, 6);
            Assert.AreEqual(b3.Length, b4.Length, "b4 length");
            for(int i = 0; i < b4.Length; i++)
            {
                Assert.AreEqual(b3[i], b4[i], "b4 element " + i);
            }
        }
        private const int REP = 1000;
        private static Random rng = new Random();
        [Test]
        public void TestPackedBCDMany()
        {
            for(int i = 0; i < REP; i++)
            {
                int v = rng.Next();
                String s = String.Format("{0:D}.{1:D2}", v / 100, Math.Abs(v % 100));
                Assert.AreEqual(decimal.Parse(s), BCDUtil.DecodePackedBCD(BCDUtil.EncodePackedBCD(decimal.Parse(s), 2, 10), 2), "packed BCD " + s);
            }
        }
        [Test]
        public void TestZonedBCDMany()
        {
            for(int i = 0; i < REP; i++)
            {
                int v = rng.Next();
                String s = String.Format("{0:D}.{1:D2}", v / 100, Math.Abs(v % 100));
                Assert.AreEqual(decimal.Parse(s), BCDUtil.DecodeZonedBCD(BCDUtil.EncodeZonedBCD(decimal.Parse(s), BCDUtil.EBCDIC, 2, 10), BCDUtil.EBCDIC, 2), "zoned BCD " + s);
            }
        }
    }
    [TestFixture]
    public class TestTimeUtil
    {
        [Test]
        public void TestJavaTimeBasic()
        {
            DateTime dt = (new DateTime(1970, 1, 1, 0, 0, 0, 1, DateTimeKind.Utc)).ToLocalTime();
            long t = TimeUtil.ToJavaTime(dt);
            Assert.AreEqual(1, t, "to Java");
            DateTime dt2 = TimeUtil.FromJavaTime(t);
            Assert.AreEqual(dt.Ticks, dt2.Ticks, "from Java");
        }
        [Test]
        public void TestUnixTimeBasic()
        {
            DateTime dt = (new DateTime(1970, 1, 1, 0, 0, 1, 0, DateTimeKind.Utc)).ToLocalTime();
            int t = TimeUtil.ToUnixTime(dt);
            Assert.AreEqual(1, t, "to Unix");
            DateTime dt2 = TimeUtil.FromUnixTime(t);
            Assert.AreEqual(dt.Ticks, dt2.Ticks, "from Unix");
        }
        [Test]
        public void TestVMSTimeBasic()
        {
            DateTime dt = (new DateTime(1858, 11, 17, 0, 0, 0, 1, DateTimeKind.Utc)).ToLocalTime();
            long t = TimeUtil.ToVMSTime(dt);
            Assert.AreEqual(10000 + TimeZone.CurrentTimeZone.GetUtcOffset(dt).Ticks, t, "to VMS");
            DateTime dt2 = TimeUtil.FromVMSTime(t);
            Assert.AreEqual(dt.Ticks, dt2.Ticks, "from VMS");
        }
        private const int REP = 1000;
        private static Random rng = new Random();
        [Test]
        public void TestJavaTimeMany()
        {
            for(int i = 0; i < REP; i++)
            {
                long v = rng.Next();
                Assert.AreEqual(v, TimeUtil.ToJavaTime(TimeUtil.FromJavaTime(v)), "Java time " + v);
            }
        }
        [Test]
        public void TestUnixTimeMany()
        {
            for(int i = 0; i < REP; i++)
            {
                int v = rng.Next();
                Assert.AreEqual(v, TimeUtil.ToUnixTime(TimeUtil.FromUnixTime(v)), "Unix time " + v);
            }
        }
        [Test]
        public void testVMSTimeMany()
        {
            for(int i = 0; i < REP; i++)
            {
                long v = rng.Next();
                v = Math.Abs(v);
                v = (v / 10000) * 10000;
                Assert.AreEqual(v, TimeUtil.ToVMSTime(TimeUtil.FromVMSTime(v)), "VMS time " + v);
            }
        }
    }
    [TestFixture]
    public class TestVAXFloatUtil
    {
        private const int N = 1000;
        private static Random rng = new Random();
        [Test]
        public void TestVAX2IEEE()
        {
            uint f = 0x70A44245U;
            ulong g = 0xBE771A9FDD2F407EUL;
            ulong d = 0xF3B8D4FDE97843F6UL;
            uint s = 0x414570A4U;
            ulong t = 0x405EDD2F1A9FBE77UL;
            Assert.AreEqual(s, VAXFloatUtil.F2S(f), "f2s");
            Assert.AreEqual(t, VAXFloatUtil.G2T(g), "g2t");
            Assert.AreEqual(t, VAXFloatUtil.D2T(d), "d2t");
        }
        [Test]
        public void TestIEEE2VAX()
        {
            uint f = 0x70A44245U;
            ulong g = 0xBE771A9FDD2F407EUL;
            ulong d = 0xF3B8D4FDE97843F6UL;
            uint s = 0x414570A4U;
            ulong t = 0x405EDD2F1A9FBE77UL;
            Assert.AreEqual(f, VAXFloatUtil.S2F(s), "s2f");
            Assert.AreEqual(g, VAXFloatUtil.T2G(t), "t2g");
            Assert.AreEqual(d, VAXFloatUtil.T2D(t), "t2d");
        }
        [Test]
        public void TestMulti32()
        {
            for(int i = 0; i < N; i++)
            {
                float xs = (float)rng.NextDouble();
                uint s = BitConverter.ToUInt32(BitConverter.GetBytes(xs), 0);
                uint f = VAXFloatUtil.S2F(s);
                uint z = VAXFloatUtil.F2S(f);
                float xz = BitConverter.ToSingle(BitConverter.GetBytes(z), 0);
                Assert.AreEqual(xs, xz, 0.0000001, "s2f2s");
            }
        }
        [Test]
        public void TestMulti64()
        {
            for(int i = 0; i < N; i++)
            {
                double xt = rng.NextDouble();
                ulong t = BitConverter.ToUInt64(BitConverter.GetBytes(xt), 0);
                ulong g = VAXFloatUtil.T2G(t);
                ulong z = VAXFloatUtil.G2T(g);
                double xz = BitConverter.ToDouble(BitConverter.GetBytes(z), 0);
                Assert.AreEqual(xt, xz, 0.0000001, "t2g2t");
            }
        }
    }
    [TestFixture]
    public class TestDictionaryHelper
    {
    	[Test]
	    public void TestFromDictionary()
	    {
			IDictionary<string,object> d = new Dictionary<string,object>();
			d["iv"] = 258;
			d["xv"] = 123.456;
			d["sv"] = "ABC     ";
			Data o = DictionaryHelper.FromDictionary<Data>(typeof(Data), d);
			Assert.AreEqual(258, o.Iv, "iv");
			Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
			Assert.AreEqual("ABC     ", o.Sv, "sv");
	    }
    	[Test]
    	public void TestToDictionary()
    	{
			Data o = new Data();
			o.Iv = 258;
			o.Xv = 123.456;
			o.Sv = "ABC     ";
			IDictionary<string,object> d = DictionaryHelper.ToDictionary<Data>(o);
			Assert.AreEqual(258, (int)d["iv"], "iv");
			Assert.AreEqual(123.456, (double)d["xv"], 0.0005, "xv");
			Assert.AreEqual("ABC     ", (string)d["sv"], "sv");
    	}
	    [Test]
	    public void TestComplexFromDictionary()
	    {
			IDictionary<string,object> d = new Dictionary<string,object>();
			d["iv"] = 1;
			d["sv"] = "A";
			IDictionary<string,object> d1 = new Dictionary<string,object>();
			d1["iv"] = 2;
			d1["sv"] = "B";
			IDictionary<string,object> d2 = new Dictionary<string,object>();
			d2["iv"] = 3;
			d2["sv"] = "C";
			IDictionary<string,object> d3 = new Dictionary<string,object>();
			d3["iv"] = 4;
			d3["sv"] = "D";
			IList<object> lst = new List<object>();
			lst.Add(d1);
			lst.Add(d2);
			lst.Add(d3);
			d["sub"] = lst;
			ComplexData o = DictionaryHelper.FromDictionary<ComplexData>(typeof(ComplexData), d);
			Assert.AreEqual(1, o.Iv, "iv");
			Assert.AreEqual("A", o.Sv, "sv");
			Assert.AreEqual(2, o.Sub[0].Iv, "sub[0].iv");
			Assert.AreEqual("B", o.Sub[0].Sv, "sub[0].sv");
			Assert.AreEqual(3, o.Sub[1].Iv, "sub[1].iv");
			Assert.AreEqual("C", o.Sub[1].Sv, "sub[1].sv");
			Assert.AreEqual(4, o.Sub[2].Iv, "sub[2].iv");
			Assert.AreEqual("D", o.Sub[2].Sv, "sub[2].sv");
	    }
	    [Test]
	    public void TestComplexToDictionary()
	    {
    		ComplexData o = new ComplexData();
    		o.Iv = 1;
    		o.Sv = "A";
    		ComplexSubData o1 = new ComplexSubData();
    		o1.Iv = 2;
    		o1.Sv = "B";
    		o.sub[0] = o1;
    		ComplexSubData o2 = new ComplexSubData();
    		o2.Iv = 3;
    		o2.Sv = "C";
    		o.sub[1] = o2;
    		ComplexSubData o3 = new ComplexSubData();
    		o3.Iv = 4;
    		o3.Sv = "D";
    		o.sub[2] = o3;
    		IDictionary<string,object> d = DictionaryHelper.ToDictionary<ComplexData>(o);
    		Assert.AreEqual(1, d["iv"], "iv");
    		Assert.AreEqual("A", d["sv"], "sv");
	    }
    }
}
