﻿/*
 * Copyright 2010 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;

using NUnit.Framework;
using log4net;
using log4net.Config;
using log4net.Core;
using log4net.Appender;
using log4net.Layout;
using log4net.Repository.Hierarchy;

namespace Vajhoej.Record.Test
{
    [TestFixture]
    public class TestStructReader
    {
        private const int N = 100000;
        [Test]
        public void TestReadSimple()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                Data o = sr.Read<Data>(typeof(Data));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
                Assert.AreEqual("ABC     ", o.Sv, "sv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadDictionary()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                IDictionary<string,object> d = sr.ReadDictionary<Data>(typeof(Data));
				Assert.AreEqual(258, (int)d["iv"], "iv");
				Assert.AreEqual(123.456, (double)d["xv"], 0.0005, "xv");
				Assert.AreEqual("ABC     ", (string)d["sv"], "sv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadMulti()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x03, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                SuperData o1 = sr.Read<SuperData>(typeof(SuperData));
                Assert.AreEqual(1, o1.Id, "id");
                Assert.AreEqual(1, o1.Typ, "typ");
                SubDataOne o1m = (SubDataOne)o1;
                Assert.AreEqual(123.456, o1m.X, 0.0005, "x");
                SuperData o2 = sr.Read<SuperData>(typeof(SuperData));
                Assert.AreEqual(3, o2.Id, "id");
                Assert.AreEqual(2, o2.Typ, "typ");
                SubDataTwo o2m = (SubDataTwo)o2;
                Assert.AreEqual("ABC         ", o2m.S, "s");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadMultiAlt()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x03, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                SubDataOne o1m = sr.Read<SubDataOne>(typeof(SubDataOne));
                Assert.AreEqual(1, o1m.Id, "id");
                Assert.AreEqual(1, o1m.Typ, "typ");
                Assert.AreEqual(123.456, o1m.X, 0.0005, "x");
                SubDataTwo o2m = sr.Read<SubDataTwo>(typeof(SubDataTwo));
                Assert.AreEqual(3, o2m.Id, "id");
                Assert.AreEqual(2, o2m.Typ, "typ");
                Assert.AreEqual("ABC         ", o2m.S, "s");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadMultiPad()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40, 0x00, 0x00, 0x00, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                SuperDataPad o1 = sr.Read<SuperDataPad>(typeof(SuperDataPad));
                Assert.AreEqual(1, o1.Id, "id");
                Assert.AreEqual(1, o1.Typ, "typ");
                SubDataOnePad o1m = (SubDataOnePad)o1;
                Assert.AreEqual(123.456, o1m.X, 0.0005, "x");
                SuperDataPad o2 = sr.Read<SuperDataPad>(typeof(SuperDataPad));
                Assert.AreEqual(3, o2.Id, "id");
                Assert.AreEqual(2, o2.Typ, "typ");
                SubDataTwoPad o2m = (SubDataTwoPad)o2;
                Assert.AreEqual("ABC         ", o2m.S, "s");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadPerf()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            try
            {
                long t1 = DateTime.Now.Ticks / 10000;
                StructReader sr = new StructReader(bm);
                for(int i = 0; i < N; i++)
                {
                    Data o = sr.Read<Data>(typeof(Data));
                    Assert.AreEqual(258, o.Iv, "iv #" + i);
                }
                long t2 = DateTime.Now.Ticks / 10000;
                Assert.IsTrue(t2-t1 < 0.1*N, "time");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadCache()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructInfoCache.Instance.Reset();
                StructReader sr = new StructReader(b);
                Data o = sr.Read<Data>(typeof(Data));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.IsTrue(StructInfoCache.Instance.HitRate < 0.001, "hit rate");
                for(int i = 0; i < 9; i++)
                {
                    StructReader sr2 = new StructReader(b);
                    Data o2 = sr2.Read<Data>(typeof(Data));
                    Assert.AreEqual(258, o2.Iv, "iv");
                }
                Assert.IsTrue(StructInfoCache.Instance.HitRate > 0.899, "hit rate");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadBoolean()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };
            try
            {
                StructReader sr = new StructReader(b);
                BooleanData o = sr.Read<BooleanData>(typeof(BooleanData));
                Assert.AreEqual(true, o.B1, "b1");
                Assert.AreEqual(false, o.B2, "b2");
                Assert.AreEqual(true, o.B3, "b3");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadBits()
        {
            byte[] b = { 0x01, 0x23, 0x04, (byte)0x80, 0x02 };
            try
            {
                StructReader sr = new StructReader(b);
                BitData o = sr.Read<BitData>(typeof(BitData));
                Assert.AreEqual(1, o.I1, "i1");
                Assert.AreEqual(2, o.N1, "n1");
                Assert.AreEqual(3, o.N2, "n2");
                Assert.AreEqual(4, o.I2, "i3");
                Assert.AreEqual(1, o.Z1, "z1");
                Assert.AreEqual(2, o.Z2, "z2");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadTime()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00 };
            try {
                StructReader sr = new StructReader(b);
                TimeData o = sr.Read<TimeData>(typeof(TimeData));
                o.T1 = o.T1.ToUniversalTime();
                Assert.AreEqual(1970, o.T1.Year, "t1 year");
                Assert.AreEqual(1, o.T1.Month, "t1 month");
                Assert.AreEqual(1, o.T1.Day, "t1 day");
                Assert.AreEqual(0, o.T1.Hour, "t1 hour");
                Assert.AreEqual(0, o.T1.Minute, "t1 minute");
                Assert.AreEqual(0, o.T1.Second, "t1 second");
                Assert.AreEqual(1, o.T1.Millisecond, "t1 millisecond");
                o.T2 = o.T2.ToUniversalTime();
                Assert.AreEqual(1858, o.T2.Year, "t2 year");
                Assert.AreEqual(11, o.T2.Month, "t2 month");
                Assert.AreEqual(17, o.T2.Day, "t2 day");
                Assert.AreEqual(5, o.T2.Hour, "t2 hour");
                Assert.AreEqual(0, o.T2.Minute, "t2 minute");
                Assert.AreEqual(0, o.T2.Second, "t2 second");
                Assert.AreEqual(0, o.T2.Millisecond, "t2 millisecond");
                o.T3 = o.T3.ToUniversalTime();
                Assert.AreEqual(1970, o.T3.Year, "t3 year");
                Assert.AreEqual(1, o.T3.Month, "t3 month");
                Assert.AreEqual(1, o.T3.Day, "t3 day");
                Assert.AreEqual(0, o.T3.Hour, "t3 hour");
                Assert.AreEqual(0, o.T3.Minute, "t3 minute");
                Assert.AreEqual(1, o.T3.Second, "t3 second");
                Assert.AreEqual(0, o.T3.Millisecond, "t3 millisecond");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadBCD()
        {
            byte[] b = { 0x12, 0x34, 0x56, 0x7D,
                         (byte)0xF1, (byte)0xF2, (byte)0xF7, (byte)0xF9, (byte)0xF5, (byte)0xC0  };
            try
            {
                StructReader sr = new StructReader(b);
                BCDData o = sr.Read<BCDData>(typeof(BCDData));
                Assert.AreEqual(decimal.Parse("-12345.67", CultureInfo.InvariantCulture), o.V1, "v1");
                Assert.AreEqual(decimal.Parse("1279.50", CultureInfo.InvariantCulture), o.V2, "v2");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadVaxFloat()
        {
            byte[] b = { 0x45, 0x42, (byte)0xA4, 0x70,
                         0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE };
            try
            {
                StructReader sr = new StructReader(b);
                VaxFloatData o = sr.Read<VaxFloatData>(typeof(VaxFloatData));
                Assert.AreEqual(12.34f, o.V1, 0.0001f, "v1");
                Assert.AreEqual(123.456, o.V2, 0.00001, "v2");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadVaxFloatArray() {
            byte[] b = { (byte)0x99, 0x40, (byte)0x9A, (byte)0x99,
            		     0x45, 0x42, (byte)0xA4, 0x70,
            		     (byte)0xF6, 0x43, 0x79, (byte)0xE9,
            		     0x00, 0x00, 0x00, 0x00,
            		     0x00, (byte)0xC0, 0x00, 0x00,
            		     0x20, (byte)0xC1, 0x00, 0x00,
            		     (byte)0xDB, 0x2F, (byte)0xFF, (byte)0xE6,
            		     0x15, 0x51, (byte)0xF9, 0x02,
            		     0x13, 0x40, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33,
            		     0x48, 0x40, 0x14, (byte)0xAE, (byte)0xE1, 0x7A, (byte)0xAE, 0x47,
            		     0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE,
            		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x24, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     (byte)0xFB, 0x3D, (byte)0xDF, 0x7C, (byte)0xD7, (byte)0xD9, (byte)0xBB, (byte)0xBD,
            		     0x22, 0x42, 0x5F, (byte)0xA0, 0x00, 0x20, 0x00, 0x00,
            		     (byte)0x99, 0x40, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x98, (byte)0x99,
            		     0x45, 0x42, (byte)0xA3, 0x70, 0x0A, (byte)0xD7, 0x70, 0x3D,
            		     (byte)0xF6, 0x43, 0x78, (byte)0xE9, (byte)0xFD, (byte)0xD4, (byte)0xB8, (byte)0xF3,
            		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x20, (byte)0xC1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     (byte)0xDB, 0x2F, (byte)0xFE, (byte)0xE6, (byte)0xBD, (byte)0xCE, (byte)0xD8, (byte)0xED,
            		     0x15, 0x51, (byte)0xF9, 0x02, 0x00, 0x00, 0x00, 0x00 };
            float[] x1 = { 1.2f, 12.34f, 123.456f, 0.0f, -0.5f, -2.5f, 0.0000000001f, 10000000000.0f };
            double[] x2 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
            double[] x3 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
            try
            {
                StructReader sr = new StructReader(b);
                VaxFloatArrayData o = sr.Read<VaxFloatArrayData>(typeof(VaxFloatArrayData));
                float[] v1 = o.V1;
                double[] v2 = o.V2;
                double[] v3 = o.V3;
                for(int i = 0; i < 8; i++)
                {
                    Assert.AreEqual(x1[i], v1[i], Math.Max(x1[i] * 0.0000001f, 0.1e-30f), "v1[" + i + "]");
                    Assert.AreEqual(x2[i], v2[i], Math.Max(x2[i] * 0.00000001, 0.1e-30), "v2[" + i + "]");
                    Assert.AreEqual(x2[i], v3[i], Math.Max(x3[i] * 0.00000001, 0.1e-30), "v3[" + i + "]");
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadUnsigned()
        {
            byte[] b = { (byte)0xFF,
                         (byte)0xFF, (byte)0xFF,
                         (byte)0xFF, (byte)0xFF,(byte)0xFF, (byte)0xFF };
            try
            {
                StructReader sr = new StructReader(b);
                UnsignedData o = sr.Read<UnsignedData>(typeof(UnsignedData));
                Assert.AreEqual(255, o.Ui1v, "ui1v");
                Assert.AreEqual(65535, o.Ui2v, "ui2v");
                Assert.AreEqual(4294967295L, o.Ui4v, "ui4v");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadArray()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructReader sr = new StructReader(b);
                ArrayData o = sr.Read<ArrayData>(typeof(ArrayData));
                for(int ix = 0; ix < 3; ix++)
                {
                    Assert.AreEqual(258 + ix, o.GetIv(ix), "iv");
                    Assert.AreEqual(123.456, o.GetXv(ix), 0.0005, "xv");
                }
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadStructField()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x04, 0x00, 0x00, 0x00,
                         0x05, 0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                MainData o = sr.Read<MainData>(typeof(MainData));
                Assert.AreEqual(1, o.I1, "i1");
                Assert.AreEqual(2, o.I2, "i2");
                Assert.AreEqual(3, o.s.I3, "i3");
                Assert.AreEqual(4, o.s.I4, "i4");
                Assert.AreEqual(5, o.I5, "i5");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadStructAlignment0()
        {
            byte[] b = { 0x01, 0x02, 0x00, 0x03,
                         0x00, 0x00, 0x00, 0x04,
                         0x00, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                AlignData0 o = sr.Read<AlignData0>(typeof(AlignData0));
                Assert.AreEqual(1, o.I1, "i1");
                Assert.AreEqual(2, o.I2, "i2");
                Assert.AreEqual(3, o.I4, "i4");
                Assert.AreEqual(4, o.I8, "i8");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadStructAlignment1()
        {
            byte[] b = { 0x01, 0x00, 0x02, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x04, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                AlignData1 o = sr.Read<AlignData1>(typeof(AlignData1));
                Assert.AreEqual(1, o.I1, "i1");
                Assert.AreEqual(2, o.I2, "i2");
                Assert.AreEqual(3, o.I4, "i4");
                Assert.AreEqual(4, o.I8, "i8");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadStructAlignment2()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00,
                         0x04, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                AlignData2 o = sr.Read<AlignData2>(typeof(AlignData2));
                Assert.AreEqual(1, o.I1, "i1");
                Assert.AreEqual(2, o.I2, "i2");
                Assert.AreEqual(3, o.I4, "i4");
                Assert.AreEqual(4, o.I8, "i8");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadError1()
        {
            byte[] b = { 0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                ErrorData1 o = sr.Read<ErrorData1>(typeof(ErrorData1));
                Assert.Fail("Missing exception");
                o.ToString();
            }
            catch (RecordException)
            {
            }
        }
        [Test]
        public void TestReadError2()
        {
            byte[] b = { };
            try
            {
                StructReader sr = new StructReader(b);
                ErrorData2 o = sr.Read<ErrorData2>(typeof(ErrorData2));
                Assert.Fail("Missing exception");
                o.ToString();
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (ArgumentException)
            {
            }
        }
        [Test]
        public void TestReadError3()
        {
            byte[] b = { };
            try
            {
                StructReader sr = new StructReader(b);
                ErrorData3 o = sr.Read<ErrorData3>(typeof(ErrorData3));
                Assert.Fail("Missing exception");
                o.ToString();
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (ArgumentException)
            {
            }
        }
        [Test]
        public void TestReadErrorMissing()
        {
            byte[] b = { };
            try
            {
                StructReader sr = new StructReader(b);
                Data o = sr.Read<Data>(typeof(Data));
                Assert.Fail("Missing exception");
                o.ToString();
            }
            catch (RecordException)
            {
            }
        }
        [Test]
        public void TestReadStrings()
        {
            byte[] b = { 0x41, 0x42, 0x43, 0x44,
                         0x02, 0x00, 0x41, 0x42,
                         0x02, 0x00, 0x41, 0x42, 0x00, 0x00,
                         0x41, 0x42, 0x00, 0x00,
                         0x02, 0x41, 0x42,
                         0x41, 0x42, 0x020, 0x20  };
            try
            {
                StructReader sr = new StructReader(b);
                StringData o = sr.Read<StringData>(typeof(StringData));
                Assert.AreEqual("ABCD", o.S1, "s1");
                Assert.AreEqual("AB", o.S2, "s2");
                Assert.AreEqual("AB", o.S3, "s3");
                Assert.AreEqual("AB", o.S4, "s4");
                Assert.AreEqual("AB", o.S5, "s5");
                Assert.AreEqual("AB", o.S6, "s6");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadVariable()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                StructReader sr = new StructReader(b);
                LengthProvider lenpvd = delegate(object ctx, int n) { return n + 1; };
                VariableData o = sr.Read<VariableData>(typeof(VariableData), lenpvd);
                Assert.AreEqual("A", o.S1, "s1");
                Assert.AreEqual("AB", o.S2, "s2");
                Assert.AreEqual("ABC", o.S3, "s3");
                Assert.AreEqual("ABCD", o.S4, "s4");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestReadVariableLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                StructReader sr = new StructReader(b);
                VariableData o = sr.Read<VariableData>(typeof(VariableData), (ctx, n) => n + 1);
                Assert.AreEqual("A", o.S1, "s1");
                Assert.AreEqual("AB", o.S2, "s2");
                Assert.AreEqual("ABC", o.S3, "s3");
                Assert.AreEqual("ABCD", o.S4, "s4");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestReadBigEndian()
        {
            byte[] b = { 0x00, 0x00, 0x01, 0x02 };
            try
            {
                StructReader sr = new StructReader(b);
                BigEndianData o = sr.Read<BigEndianData>(typeof(BigEndianData));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadMixedEndian()
        {
            byte[] b = { 0x00, 0x01, 0x02, 0x00, 0x00, 0x03, 0x00, 0x04 };
            try
            {
                StructReader sr = new StructReader(b);
                MixedEndianData o = sr.Read<MixedEndianData>(typeof(MixedEndianData));
                Assert.AreEqual(1, o.Iv1, "iv1");
                Assert.AreEqual(2, o.Iv2, "iv2");
                Assert.AreEqual(3, o.Iv3, "iv3");
                Assert.AreEqual(4, o.Iv4, "iv4");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadEndPad1()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x01 };
            try
            {
                StructReader sr = new StructReader(b);
                EndPadData1 o = sr.Read<EndPadData1>(typeof(EndPadData1));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual(1, o.Bv, "bv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadEndPad2()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                EndPadData2 o = sr.Read<EndPadData2>(typeof(EndPadData2));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual(1, o.Bv, "bv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadEndPad3()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructReader sr = new StructReader(b);
                EndPadData3 o = sr.Read<EndPadData3>(typeof(EndPadData3));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual(0.0, o.Xv, 0.0005, "xv");
                Assert.AreEqual(1, o.Bv, "bv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadLogging()
        {
            int n1 = TestLogging(Level.Info);
            Assert.AreEqual(0, n1, "INFO log level");
            int n2 = TestLogging(Level.Debug);
            Assert.IsTrue(n2 > n1, "DEBUG log level");
        }
        private class MyLayout : LayoutSkeleton
        {
            public MyLayout()
            {
                this.IgnoresException = false;
            }
            public override void ActivateOptions()
            {
            }
            public override void Format(TextWriter tw, LoggingEvent evt)
            {
                tw.WriteLine(String.Format("{0} {1} {2} {3}",
                                           evt.TimeStamp, evt.LoggerName, evt.Level, evt.MessageObject.ToString()));
            }
        }
        private class MyAppender : AppenderSkeleton
        {
            private ILayout layout = new MyLayout();
            private TextWriter tw;
            public MyAppender(TextWriter tw)
            {
                this.tw = tw;
            }
            protected override void Append(LoggingEvent evt)
            {
                layout.Format(tw, evt);
            }
        }
        public int TestLogging(Level lvl)
        {
            BasicConfigurator.Configure();
            StringWriter lw = new StringWriter();
            Logger log = (Logger)LogManager.GetLogger("Vajhoej.Record").Logger;
            log.Level = lvl;
            log.Hierarchy.Root.RemoveAllAppenders();
            IAppender app = new MyAppender(lw);
            log.AddAppender(app);
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                Data o = sr.Read<Data>(typeof(Data));
                o.ToString();
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            log.RemoveAppender(app);
            log.Level = Level.Off;
            String logoutput = lw.ToString();
            return logoutput.Length;
        }
        [Test]
        public void TestReadValueType()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructReader sr = new StructReader(b);
                ValueTypeData o = sr.Read<ValueTypeData>(typeof(ValueTypeData));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
                Assert.AreEqual("ABC     ", o.Sv, "sv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadGeneralInt()
        {
            byte[] b = { 0x03, 0x02, 0x01, 0x05, 0x04, 0x03, 0x02, 0x01 };
            try
            {
                StructReader sr = new StructReader(b);
                GeneralIntData o = sr.Read<GeneralIntData>(typeof(GeneralIntData));
                Assert.AreEqual(0x0000000000010203L, o.V1, "v1");
                Assert.AreEqual(0x0000000102030405L, o.V2, "v2");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadRemainingString()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x41, 0x42, 0x43 };
            try
            {
                StructReader sr = new StructReader(b);
                RemainingStringData o = sr.Read<RemainingStringData>(typeof(RemainingStringData));
                Assert.AreEqual(258, o.Iv, "iv");
                Assert.AreEqual("ABC", o.Sv, "sv");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadMultiConvert()
        {
            byte[] b = { 0x31, 0x41,
            		     0x32, 0x42, 0x42 };
            try
            {
                ConvertSelector inf = delegate(object o) { return int.Parse((String)o); };
                StructReader sr = new StructReader(b);
                SuperDataConvert o1 = sr.Read<SuperDataConvert>(typeof(SuperDataConvert), null, null, null, inf);
                Assert.AreEqual("1", o1.Typ, "typ");
                SubDataOneConvert o1m = (SubDataOneConvert)o1;
                Assert.AreEqual("A", o1m.A, "a");
                SuperDataConvert o2 = sr.Read<SuperDataConvert>(typeof(SuperDataConvert), null, null, null, inf);
                Assert.AreEqual("2", o2.Typ, "typ");
                SubDataTwoConvert o2m = (SubDataTwoConvert)o2;
                Assert.AreEqual("BB", o2m.B, "b");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestReadMultiConvertLambda()
        {
            byte[] b = { 0x31, 0x41,
            		     0x32, 0x42, 0x42 };
            try
            {
                StructReader sr = new StructReader(b);
                SuperDataConvert o1 = sr.Read<SuperDataConvert>(typeof(SuperDataConvert), null, null, null, o => int.Parse((String)o));
                Assert.AreEqual("1", o1.Typ, "typ");
                SubDataOneConvert o1m = (SubDataOneConvert)o1;
                Assert.AreEqual("A", o1m.A, "a");
                SuperDataConvert o2 = sr.Read<SuperDataConvert>(typeof(SuperDataConvert), null, null, null, o => int.Parse((String)o));
                Assert.AreEqual("2", o2.Typ, "typ");
                SubDataTwoConvert o2m = (SubDataTwoConvert)o2;
                Assert.AreEqual("BB", o2m.B, "b");
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestReadVarArray()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x05, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                ElementsProvider pvd2 = delegate(object ctx, int n) { return 4; };
                StructReader sr = new StructReader(b);
                VarArrayData o = sr.Read<VarArrayData>(typeof(VarArrayData), null, null, pvd2);
                for(int ix = 0; ix < 4; ix++)
                {
                    Assert.AreEqual(258 + ix, o.GetIv(ix), "iv");
                    Assert.AreEqual(123.456, o.GetXv(ix), 0.0005, "xv");
                }
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestReadVarArrayLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x05, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructReader sr = new StructReader(b);
                VarArrayData o = sr.Read<VarArrayData>(typeof(VarArrayData), null, null, (ctx,n) => 4);
                for(int ix = 0; ix < 4; ix++)
                {
                    Assert.AreEqual(258 + ix, o.GetIv(ix), "iv");
                    Assert.AreEqual(123.456, o.GetXv(ix), 0.0005, "xv");
                }
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestReadVarArrayWithSizeAdjust()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x05, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                ElementsProvider pvd2 = delegate(object ctx, int n) { return 4; };
                StructReader sr = new StructReader(b);
                VarArrayData o = sr.Read<VarArrayData>(typeof(VarArrayData), null, null, pvd2);
                for(int ix = 0; ix < 4; ix++)
                {
                    Assert.AreEqual(258 + ix, o.GetIv(ix), "iv");
                    Assert.AreEqual(123.456, o.GetXv(ix), 0.0005, "xv");
                }
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestReadVarArrayWithSizeAdjustLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x05, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructReader sr = new StructReader(b);
                VarArrayData o = sr.Read<VarArrayData>(typeof(VarArrayData), null, null, (ctx,n) => 4);
                for(int ix = 0; ix < 4; ix++)
                {
                    Assert.AreEqual(258 + ix, o.GetIv(ix), "iv");
                    Assert.AreEqual(123.456, o.GetXv(ix), 0.0005, "xv");
                }
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestReadCountArray()
        {
            byte[] b = { 0x04,
                         0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x05, 0x01, 0x00, 0x00,
                         0x04, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                ElementsProvider pvd2 = delegate(object ctx, int n) { return 4; };
                StructReader sr = new StructReader(b);
                CountArrayData o = sr.Read<CountArrayData>(typeof(CountArrayData));
                for(int ix = 0; ix < 4; ix++)
                {
                    Assert.AreEqual(258 + ix, o.GetIv(ix), "iv");
                    Assert.AreEqual(123.456, o.GetXv(ix), 0.0005, "xv");
                }
                Assert.AreEqual(false, sr.More, "more");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
    }
    [TestFixture]
    public class TestStructWriter
    {
        private const int N = 100000;
        [Test]
        public void TestWriteSimple()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
                Data o = new Data();
                o.Iv = 258;
                o.Xv = 123.456;
                o.Sv = "ABC     ";
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteDictionary()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
				IDictionary<string,object> d = new Dictionary<string,object>();
				d["iv"] = 258;
				d["xv"] = 123.456;
				d["sv"] = "ABC     ";
				sw.WriteDictionary<Data>(typeof(Data), d);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteMulti()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x03, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
                SubDataOne o1 = new SubDataOne();
                o1.Id = 1;
                o1.Typ = 1;
                o1.X = 123.456;
                SuperData o1m = o1;
                sw.Write(o1m);
                SubDataTwo o2 = new SubDataTwo();
                o2.Id = 3;
                o2.Typ = 2;
                o2.S = "ABC         ";
                SuperData o2m = o2;
                sw.Write(o2m);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteMultiAlt()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x03, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
                SubDataOne o1 = new SubDataOne();
                o1.Id = 1;
                o1.Typ = 1;
                o1.X = 123.456;
                sw.Write(o1);
                SubDataTwo o2 = new SubDataTwo();
                o2.Id = 3;
                o2.Typ = 2;
                o2.S = "ABC         ";
                sw.Write(o2);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteMultiPad()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40, 0x00, 0x00, 0x00, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
                SubDataOnePad o1 = new SubDataOnePad();
                o1.Id = 1;
                o1.Typ = 1;
                o1.X = 123.456;
                SuperDataPad o1m = o1;
                sw.Write(o1m);
                SubDataTwoPad o2 = new SubDataTwoPad();
                o2.Id = 3;
                o2.Typ = 2;
                o2.S = "ABC         ";
                SuperDataPad o2m = o2;
                sw.Write(o2m);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWritePerf()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                long t1 = DateTime.Now.Ticks / 10000;
                StructWriter sw = new StructWriter(100*N);
                for(int i = 0; i < N; i++)
                {
                    Data o = new Data();
                    o.Iv = 258;
                    o.Xv = 123.456;
                    o.Sv = "ABC     ";
                    sw.Write(o);
                }
                byte[] res = sw.GetBytes();
                long t2 = DateTime.Now.Ticks / 10000;
                Assert.AreEqual(N*b.Length, res.Length, "length");
                Assert.IsTrue(t2-t1 < 0.1*N, "time");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteCache()
        {
            try
            {
                StructInfoCache.Instance.Reset();
                StructWriter sw = new StructWriter();
                Data o = new Data();
                o.Iv = 258;
                o.Xv = 123.456;
                o.Sv = "ABC     ";
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(20, res.Length, "length");
                Assert.IsTrue(StructInfoCache.Instance.HitRate < 0.001, "hit rate");
                for(int i = 0; i < 9; i++)
                {
                    StructWriter sw2 = new StructWriter();
                    Data o2 = new Data();
                    o2.Iv = 258;
                    o2.Xv = 123.456;
                    o2.Sv = "ABC     ";
                    sw2.Write(o2);
                    byte[] res2 = sw.GetBytes();
                    Assert.AreEqual(20, res2.Length, "length");
                }
                Assert.IsTrue(StructInfoCache.Instance.HitRate > 0.899, "hit rate");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteBoolean()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };
            try
            {
                StructWriter sw = new StructWriter();
                BooleanData o = new BooleanData();
                o.B1 = true;
                o.B2 = false;
                o.B3 = true;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteBits()
        {
            byte[] b = { 0x01, 0x23, 0x04, (byte)0x80, 0x02 };
            try
            {
                StructWriter sw = new StructWriter();
                BitData o = new BitData();
                o.I1 = 1;
                o.N1 = 2;
                o.N2 = 3;
                o.I2 = 4;
                o.Z1 = 1;
                o.Z2 = 2;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteTime()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x10, 0x27, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                TimeData o = new TimeData();
                o.T1 = new DateTime(1970, 1, 1, 0, 0, 0, 1, DateTimeKind.Utc);
                o.T2 = new DateTime(1858, 11, 17, 0, 0, 0, 1, DateTimeKind.Utc);
                o.T3 = new DateTime(1970, 1, 1, 0, 0, 1, 0, DateTimeKind.Utc);
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteBCD()
        {
            byte[] b = { 0x12, 0x34, 0x56, 0x7D,
                        (byte)0xF1, (byte)0xF2, (byte)0xF7, (byte)0xF9, (byte)0xF5, (byte)0xC0  };
            try
            {
                StructWriter sw = new StructWriter();
                BCDData o = new BCDData();
                o.V1 = decimal.Parse("-12345.67", CultureInfo.InvariantCulture);
                o.V2 = decimal.Parse("1279.50", CultureInfo.InvariantCulture);
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteVaxFloat()
        {
            byte[] b = { 0x45, 0x42, (byte)0xA4, 0x70,
                         0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE };
            try
            {
                StructWriter sw = new StructWriter();
                VaxFloatData o = new VaxFloatData();
                o.V1  = 12.34f;
                o.V2 = 123.456;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteVaxFloatArray() {
            byte[] b = { (byte)0x99, 0x40, (byte)0x9A, (byte)0x99,
             		     0x45, 0x42, (byte)0xA4, 0x70,
       		             (byte)0xF6, 0x43, 0x79, (byte)0xE9,
       		             0x00, 0x00, 0x00, 0x00,
               		     0x00, (byte)0xC0, 0x00, 0x00,
       		             0x20, (byte)0xC1, 0x00, 0x00,
            		     (byte)0xDB, 0x2F, (byte)0xFF, (byte)0xE6,
            		     0x15, 0x51, (byte)0xF9, 0x02,
            		     0x13, 0x40, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33,
            		     0x48, 0x40, 0x14, (byte)0xAE, (byte)0xE1, 0x7A, (byte)0xAE, 0x47,
            		     0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE,
            		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x24, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     (byte)0xFB, 0x3D, (byte)0xDF, 0x7C, (byte)0xD7, (byte)0xD9, (byte)0xBB, (byte)0xBD,
             		     0x22, 0x42, 0x5F, (byte)0xA0, 0x00, 0x20, 0x00, 0x00,
             		     (byte)0x99, 0x40, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x98, (byte)0x99,
            		     0x45, 0x42, (byte)0xA3, 0x70, 0x0A, (byte)0xD7, 0x70, 0x3D,
            		     (byte)0xF6, 0x43, 0x78, (byte)0xE9, (byte)0xFD, (byte)0xD4, (byte)0xB8, (byte)0xF3,
            		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     0x20, (byte)0xC1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            		     (byte)0xDB, 0x2F, (byte)0xFE, (byte)0xE6, (byte)0xBD, (byte)0xCE, (byte)0xD8, (byte)0xED,
       		             0x15, 0x51, (byte)0xF9, 0x02, 0x00, 0x00, 0x00, 0x00 };
            float[] x1 = { 1.2f, 12.34f, 123.456f, 0.0f, -0.5f, -2.5f, 0.0000000001f, 10000000000.0f };
            double[] x2 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
            double[] x3 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
            try
            {
                StructWriter sw = new StructWriter();
                VaxFloatArrayData o = new VaxFloatArrayData();
                o.V1 = x1;
                o.V2 = x2;
                o.V3 = x3;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteUnsigned()
        {
            byte[] b = { (byte)0xFF,
                         (byte)0xFF, (byte)0xFF,
                         (byte)0xFF, (byte)0xFF,(byte)0xFF, (byte)0xFF };
            try
            {
                StructWriter sw = new StructWriter();
                UnsignedData o = new UnsignedData();
                o.Ui1v = 255;
                o.Ui2v = 65535;
                o.Ui4v = 4294967295;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteArray()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x03, 0x01, 0x00, 0x00,
                         0x04, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructWriter sw = new StructWriter();
                ArrayData o = new ArrayData();
                for(int ix = 0; ix < 3; ix++)
                {
                    o.SetIv(ix, 258 + ix);
                    o.SetXv(ix, 123.456);
                }
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteStructField()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x04, 0x00, 0x00, 0x00,
                         0x05, 0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                MainData o = new MainData();
                o.I1 = 1;
                o.I2 = 2;
                o.s = new FieldData();
                o.s.I3 = 3;
                o.s.I4 = 4;
                o.I5 = 5;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteStructAlignment0()
        {
            byte[] b = { 0x01, 0x02, 0x00, 0x03,
                         0x00, 0x00, 0x00, 0x04,
                         0x00, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                AlignData0 o = new AlignData0();
                o.I1 = 1;
                o.I2 = 2;
                o.I4 = 3;
                o.I8 = 4;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteStructAlignment1()
        {
            byte[] b = { 0x01, 0x00, 0x02, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x04, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                AlignData1 o = new AlignData1();
                o.I1 = 1;
                o.I2 = 2;
                o.I4 = 3;
                o.I8 = 4;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteStructAlignment2()
        {
            byte[] b = { 0x01, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00,
                         0x02, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00,
                         0x03, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00,
                         0x04, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                AlignData2 o = new AlignData2();
                o.I1 = 1;
                o.I2 = 2;
                o.I4 = 3;
                o.I8 = 4;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++) {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteError1()
        {
            try
            {
                StructWriter sw = new StructWriter();
                ErrorData1 o = new ErrorData1();
                o.Sv = "ABC";
                sw.Write(o);
                Assert.Fail("Missing exception");
            }
            catch (RecordException)
            {
            }
        }
        [Test]
        public void TestWriteError2()
        {
            try
            {
                StructWriter sw = new StructWriter();
                ErrorData2 o = new ErrorData2();
                o.Iv = 123;
                sw.Write(o);
                Assert.Fail("Missing exception");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (ArgumentException)
            {
            }
        }
        [Test]
        public void TestWriteError3()
        {
            try
            {
                StructWriter sw = new StructWriter();
                ErrorData3 o = new ErrorData3();
                o.Iv1 = 123;
                o.Iv2 = 456;
                sw.Write(o);
                Assert.Fail("Missing exception");
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (ArgumentException)
            {
            }
        }
        [Test]
        public void TestWriteStrings()
        {
            byte[] b = { 0x41, 0x42, 0x43, 0x44,
                         0x02, 0x00, 0x41, 0x42,
                         0x02, 0x00, 0x41, 0x42, 0x00, 0x00,
                         0x41, 0x42, 0x00, 0x00,
                         0x02, 0x41, 0x42,
                         0x41, 0x42, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
                StringData o = new StringData();
                o.S1 = "ABCD";
                o.S2 = "AB";
                o.S3 = "AB";
                o.S4 = "AB";
                o.S5 = "AB";
                o.S6 = "AB";
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
	    public void TestWriteStringsNull()
		{
	        byte[] b = { 0x20, 0x20, 0x20, 0x20,
	                     0x00, 0x00,
	                     0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	                     0x00, 0x00, 0x00, 0x00,
	                     0x00, 
	                     0x20, 0x20, 0x20, 0x20 };
	        try
	        {
	            StructWriter sw = new StructWriter();
	            StringData o = new StringData();
	            o.S1 = null;
	            o.S2 = null;
	            o.S3 = null;
	            o.S4 = null;
	            o.S5 = null;
	            o.S6 = null;
	            sw.Write(o);
	            byte[] res = sw.GetBytes();
	            Assert.AreEqual(b.Length, res.Length, "length");
	            for(int i = 0; i < res.Length; i++)
	            {
	                Assert.AreEqual(b[i], res[i], "byte " + i);
	            }
	        }
	        catch (RecordException e)
	        {
	            Assert.Fail("Unexpected exception: " + e);
	        }
	    }
        [Test]
        public void TestWriteVariable()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                StructWriter sw = new StructWriter();
                VariableData o = new VariableData();
                o.S1 = "A";
                o.S2 = "AB";
                o.S3 = "ABC";
                o.S4 = "ABCD";
                LengthProvider lenpvd = delegate(object ctx, int n) { return n + 1; };
                sw.Write(o, lenpvd);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestWriteVariableLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                StructWriter sw = new StructWriter();
                VariableData o = new VariableData();
                o.S1 = "A";
                o.S2 = "AB";
                o.S3 = "ABC";
                o.S4 = "ABCD";
                sw.Write(o, (ctx, n) => n + 1);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestWriteBigEndian()
        {
            byte[] b = { 0x00, 0x00, 0x01, 0x02 };
            try
            {
                StructWriter sw = new StructWriter();
                BigEndianData o = new BigEndianData();
                o.Iv = 258;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteMixedEndian()
        {
            byte[] b = { 0x00, 0x01, 0x02, 0x00, 0x00, 0x03, 0x00, 0x04 };
            try
            {
                StructWriter sw = new StructWriter();
                MixedEndianData o = new MixedEndianData();
                o.Iv1 = 1;
                o.Iv2 = 2;
                o.Iv3 = 3;
                o.Iv4 = 4;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteEndPad1()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x01 };
            try
            {
                StructWriter sw = new StructWriter();
                EndPadData1 o = new EndPadData1();
                o.Iv = 258;
                o.Bv = 1;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteEndPad2()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                EndPadData2 o = new EndPadData2();
                o.Iv = 258;
                o.Bv = 1;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteEndPad3()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
            try
            {
                StructWriter sw = new StructWriter();
                EndPadData3 o = new EndPadData3();
                o.Iv = 258;
                o.Xv = 0.0;
                o.Bv = 1;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteLogging()
        {
            int n1 = TestLogging(Level.Info);
            Assert.AreEqual(0, n1, "INFO log level");
            int n2 = TestLogging(Level.Debug);
            Assert.IsTrue(n2 > n1, "DEBUG log level");
        }
        private class MyLayout : LayoutSkeleton
        {
            public MyLayout()
            {
                this.IgnoresException = false;
            }
            public override void ActivateOptions()
            {
            }
            public override void Format(TextWriter tw, LoggingEvent evt)
            {
                tw.WriteLine(String.Format("{0} {1} {2} {3}",
                                           evt.TimeStamp, evt.LoggerName, evt.Level, evt.MessageObject.ToString()));
            }
        }
        private class MyAppender : AppenderSkeleton
        {
            private ILayout layout = new MyLayout();
            private TextWriter tw;
            public MyAppender(TextWriter tw)
            {
                this.tw = tw;
            }
            protected override void Append(LoggingEvent evt)
            {
                layout.Format(tw, evt);
            }
        }
        public int TestLogging(Level lvl)
        {
            BasicConfigurator.Configure();
            StringWriter lw = new StringWriter();
            Logger log = (Logger)LogManager.GetLogger("Vajhoej.Record").Logger;
            log.Level = lvl;
            log.Hierarchy.Root.RemoveAllAppenders();
            IAppender app = new MyAppender(lw);
            log.AddAppender(app);
            try
            {
                StructWriter sw = new StructWriter();
                Data o = new Data();
                o.Iv = 258;
                o.Xv = 123.456;
                o.Sv = "ABC     ";
                sw.Write(o);
                byte[] res = sw.GetBytes();
                res.ToString();
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            log.RemoveAppender(app);
            log.Level = Level.Off;
            String logoutput = lw.ToString();
            return logoutput.Length;
        }
        [Test]
        public void TestWriteValueType()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                StructWriter sw = new StructWriter();
                ValueTypeData o = new ValueTypeData();
                o.Iv = 258;
                o.Xv = 123.456;
                o.Sv = "ABC     ";
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteGeneralInt()
        {
            byte[] b = { 0x03, 0x02, 0x01, 0x05, 0x04, 0x03, 0x02, 0x01 };
            try
            {
                StructWriter sw = new StructWriter();
                GeneralIntData o = new GeneralIntData();
                o.V1 = 0x0000000000010203L;
                o.V2 = 0x0000000102030405L;
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteRemainingString()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x41, 0x42, 0x43 };
            try
            {
                StructWriter sw = new StructWriter();
                RemainingStringData o = new RemainingStringData();
                o.Iv = 258;
                o.Sv = "ABC";
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteMultiConvert()
        {
            byte[] b = { 0x31, 0x41,
       		             0x32, 0x42, 0x42 };
            try
            {
                ConvertSelector inf = delegate(object o) { return int.Parse((String)o); };
                StructWriter sw = new StructWriter();
                SubDataOneConvert o1 = new SubDataOneConvert();
                o1.Typ = "1";
                o1.A = "A";
                SuperDataConvert o1m = o1;
                sw.Write(o1m, null, null, null, inf);
                SubDataTwoConvert o2 = new SubDataTwoConvert();
                o2.Typ = "2";
                o2.B = "BB";
                SuperDataConvert o2m = o2;
                sw.Write(o2m, null, null, null, inf);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestWriteMultiConvertLambda()
        {
            byte[] b = { 0x31, 0x41,
       		             0x32, 0x42, 0x42 };
            try
            {
                StructWriter sw = new StructWriter();
                SubDataOneConvert o1 = new SubDataOneConvert();
                o1.Typ = "1";
                o1.A = "A";
                SuperDataConvert o1m = o1;
                sw.Write(o1m, null, null, null, o => int.Parse((String)o));
                SubDataTwoConvert o2 = new SubDataTwoConvert();
                o2.Typ = "2";
                o2.B = "BB";
                SuperDataConvert o2m = o2;
                sw.Write(o2m, null, null, null, o => int.Parse((String)o));
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestWriteVarArray()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
    	                 0x03, 0x01, 0x00, 0x00,
    	                 0x04, 0x01, 0x00, 0x00,
    	                 0x05, 0x01, 0x00, 0x00,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructWriter sw = new StructWriter();
                VarArrayData o = new VarArrayData();
                for(int ix = 0; ix < 4; ix++)
                {
                    o.SetIv(ix, 258 + ix);
                    o.SetXv(ix, 123.456);
                }
                ElementsProvider pvd2 = delegate(object oo, int n) { return 4; };
                sw.Write(o, null, null, pvd2);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35        
        [Test]
        public void TestWriteVarArrayLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
    	                 0x03, 0x01, 0x00, 0x00,
    	                 0x04, 0x01, 0x00, 0x00,
    	                 0x05, 0x01, 0x00, 0x00,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructWriter sw = new StructWriter();
                VarArrayData o = new VarArrayData();
                for(int ix = 0; ix < 4; ix++)
                {
                    o.SetIv(ix, 258 + ix);
                    o.SetXv(ix, 123.456);
                }
                sw.Write(o, null, null, (ctx,n) => 4);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestWriteCountArray()
        {
            byte[] b = { 0x04,
                         0x02, 0x01, 0x00, 0x00,
    	                 0x03, 0x01, 0x00, 0x00,
    	                 0x04, 0x01, 0x00, 0x00,
    	                 0x05, 0x01, 0x00, 0x00,
    	                 0x04, 0x00,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    	                 0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
            try
            {
                StructWriter sw = new StructWriter();
                CountArrayData o = new CountArrayData();
                for(int ix = 0; ix < 4; ix++)
                {
                    o.SetIv(ix, 258 + ix);
                    o.SetXv(ix, 123.456);
                }
                sw.Write(o);
                byte[] res = sw.GetBytes();
                Assert.AreEqual(b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
    }
    [TestFixture]
    public class TestStructReaderStream
    {
        [Test]
    	public void TestSmall()
    	{
    		byte[] b = { 0x02, 0x01, 0x00, 0x00,
    			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
    		try
    		{
    			StructReaderStream srs = new StructReaderStream(new MemoryStream(b), 20, 100);
    			Data o = srs.Read<Data>(typeof(Data));
    			Assert.AreEqual(258, o.Iv, "iv");
    			Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
    			Assert.AreEqual("ABC     ", o.Sv, "sv");
                Assert.AreEqual(false, srs.More, "more");
    		}
    		catch (RecordException e)
    		{
    			Assert.Fail("Unexpected exception: " + e);
    		}
    		catch (IOException e)
    		{
    			Assert.Fail("Unexpected exception: " + e);
    		}
    	}
        [Test]
    	public void TestBig()
    	{
    		byte[] b = { 0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20,
    			         0x02, 0x01, 0x00, 0x00,
    		             0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
    			         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
    		try
    		{
    			StructReaderStream srs = new StructReaderStream(new MemoryStream(b), 20, 5);
    			int n = 0;
    			while(srs.More)
    			{
    			    Data o = srs.Read<Data>(typeof(Data));
    				n++;
        			Assert.AreEqual(258, o.Iv, "iv");
        			Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
        			Assert.AreEqual("ABC     ", o.Sv, "sv");
    			}
                Assert.AreEqual(false, srs.More, "more");
    			Assert.AreEqual(8, n, "n");
    		}
    		catch (RecordException e)
    		{
    			Assert.Fail("Unexpected exception: " + e);
    		}
    		catch (IOException e)
    		{
    			Assert.Fail("Unexpected exception: " + e);
    		}
        }
	}
}
